/*
 * Copyright (c) 2001-2004 Bertrand Petit
 *
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file audio/mp3/mp3file.c
 */

#include <string.h>
#include <unistd.h>
#include <id3tag.h>

#include "nomad.h"

#include <audio/mp3/driver.h>
#include <audio/output/output.h>

/**
 * @brief Test if an opened file is an MP3 file.
 */
static int
mp3_match(FILE *fp)
{
	unsigned char buf[3];

	if (fread(buf, sizeof buf, 1, fp) != 1)
		return (-1);
	
	/* Match 1: first twelve bits high */
	if (buf[0] == 0xff && (buf[1] & 0xf0) == 0xf0)
		return (0);
	
	/* Match 2: ID3 header */
	if (buf[0] == 'I' && buf[1] == 'D' && buf[2] == '3')
		return (0);

	return (-1);
}

/**
 * @brief Read the ID3 tag from an MP3 file.
 */
static void
mp3_readtags(struct audio_file *fd)
{
	int tmpfd, i;
	off_t orig_off;
	struct id3_file *id3f;
	struct id3_tag *tag;
	id3_ucs4_t const *str;
	char **dst = NULL;

	/* Duplicate the current filehandle */
	tmpfd = dup(fileno(fd->fp));
	if (tmpfd < 0)
		return;
	
	/* Backing up the original offset */
	orig_off = lseek(tmpfd, 0, SEEK_CUR);

	id3f = id3_file_fdopen(tmpfd, ID3_FILE_MODE_READONLY);
	if (id3f == NULL)
		goto bad;
	
	/* Here we go */
	tag = id3_file_tag(id3f);
	if (tag == NULL)
		goto done;

	for (i = 0; i < tag->nframes; i++) {
		if (strncmp("TPE1", tag->frames[i]->id, 4) == 0) {
			dst = &fd->tag.artist;
		} else if (strncmp("TIT2", tag->frames[i]->id, 4) == 0) {
			dst = &fd->tag.title;
		} else if (strncmp("TALB", tag->frames[i]->id, 4) == 0) {
			dst = &fd->tag.album;
		} else {
			continue;
		}
		if (id3_field_getnstrings(&tag->frames[i]->fields[1]) != 0) {
			str = id3_field_getstrings(&tag->frames[i]->fields[1], 0);
			if (str != NULL)
				*dst = (char*)id3_ucs4_latin1duplicate(str);
		}
	}
	
done:
	id3_file_close(id3f);
bad:
	lseek(tmpfd, orig_off, SEEK_SET);
	close(tmpfd);
}

int
mp3_open(struct audio_file *fd)
{
	struct nomad *nf;
	struct nomad_info *ninf;

	/* Don't match other files */
	if (mp3_match(fd->fp) != 0)
		return (-1);
	rewind(fd->fp);

	if ((nf = nomad_open(fd->fp)) == NULL)
		return (-1);
	
	ninf = nomad_info(nf);

	fd->srate = ninf->sample_rate;
	fd->channels = ninf->channels;
	fd->time_len = nomad_time_total(nf);

	fd->drv_data = nf;

	mp3_readtags(fd);

	return (0);
}

void
mp3_close(struct audio_file *fd)
{
	struct nomad *nf = fd->drv_data;

	nomad_close(nf);
	fd->fp = NULL;
}

size_t
mp3_read(struct audio_file *fd, void *buf)
{
	struct nomad *nf = fd->drv_data;
	size_t ret;

	ret = nomad_read(nf, buf, AUDIO_OUTPUT_BUFLEN);
	fd->time_cur = nomad_time_tell(nf);
	return (ret);
}

void
mp3_seek(struct audio_file *fd, double len)
{
	struct nomad *nf = fd->drv_data;

	/* Nomad only supports absolute seeks */
	nomad_time_seek(nf, nomad_time_tell(nf) + len);
	fd->time_cur = nomad_time_tell(nf);
}
